import logging
import random

logger = logging.getLogger(__name__)

import h5py
import matplotlib.pyplot as plt
import numpy
import scipy.fft
import scipy.signal
import tqdm
# import opt_einsum

import numpy_popcount


AES_SBOX = numpy.array([
  # 0     1    2      3     4    5     6     7      8    9     A      B    C     D     E     F
  0x63, 0x7c, 0x77, 0x7b, 0xf2, 0x6b, 0x6f, 0xc5, 0x30, 0x01, 0x67, 0x2b, 0xfe, 0xd7, 0xab, 0x76,
  0xca, 0x82, 0xc9, 0x7d, 0xfa, 0x59, 0x47, 0xf0, 0xad, 0xd4, 0xa2, 0xaf, 0x9c, 0xa4, 0x72, 0xc0,
  0xb7, 0xfd, 0x93, 0x26, 0x36, 0x3f, 0xf7, 0xcc, 0x34, 0xa5, 0xe5, 0xf1, 0x71, 0xd8, 0x31, 0x15,
  0x04, 0xc7, 0x23, 0xc3, 0x18, 0x96, 0x05, 0x9a, 0x07, 0x12, 0x80, 0xe2, 0xeb, 0x27, 0xb2, 0x75,
  0x09, 0x83, 0x2c, 0x1a, 0x1b, 0x6e, 0x5a, 0xa0, 0x52, 0x3b, 0xd6, 0xb3, 0x29, 0xe3, 0x2f, 0x84,
  0x53, 0xd1, 0x00, 0xed, 0x20, 0xfc, 0xb1, 0x5b, 0x6a, 0xcb, 0xbe, 0x39, 0x4a, 0x4c, 0x58, 0xcf,
  0xd0, 0xef, 0xaa, 0xfb, 0x43, 0x4d, 0x33, 0x85, 0x45, 0xf9, 0x02, 0x7f, 0x50, 0x3c, 0x9f, 0xa8,
  0x51, 0xa3, 0x40, 0x8f, 0x92, 0x9d, 0x38, 0xf5, 0xbc, 0xb6, 0xda, 0x21, 0x10, 0xff, 0xf3, 0xd2,
  0xcd, 0x0c, 0x13, 0xec, 0x5f, 0x97, 0x44, 0x17, 0xc4, 0xa7, 0x7e, 0x3d, 0x64, 0x5d, 0x19, 0x73,
  0x60, 0x81, 0x4f, 0xdc, 0x22, 0x2a, 0x90, 0x88, 0x46, 0xee, 0xb8, 0x14, 0xde, 0x5e, 0x0b, 0xdb,
  0xe0, 0x32, 0x3a, 0x0a, 0x49, 0x06, 0x24, 0x5c, 0xc2, 0xd3, 0xac, 0x62, 0x91, 0x95, 0xe4, 0x79,
  0xe7, 0xc8, 0x37, 0x6d, 0x8d, 0xd5, 0x4e, 0xa9, 0x6c, 0x56, 0xf4, 0xea, 0x65, 0x7a, 0xae, 0x08,
  0xba, 0x78, 0x25, 0x2e, 0x1c, 0xa6, 0xb4, 0xc6, 0xe8, 0xdd, 0x74, 0x1f, 0x4b, 0xbd, 0x8b, 0x8a,
  0x70, 0x3e, 0xb5, 0x66, 0x48, 0x03, 0xf6, 0x0e, 0x61, 0x35, 0x57, 0xb9, 0x86, 0xc1, 0x1d, 0x9e,
  0xe1, 0xf8, 0x98, 0x11, 0x69, 0xd9, 0x8e, 0x94, 0x9b, 0x1e, 0x87, 0xe9, 0xce, 0x55, 0x28, 0xdf,
  0x8c, 0xa1, 0x89, 0x0d, 0xbf, 0xe6, 0x42, 0x68, 0x41, 0x99, 0x2d, 0x0f, 0xb0, 0x54, 0xbb, 0x16
], dtype='uint8')


def align(db, ntraces):
    logger.info("aligning traces")
    traces = db["traces"]
    reference = traces[0].astype("double")
    alignments = numpy.zeros(ntraces, "int64")
    lags = scipy.signal.correlation_lags(len(reference), len(reference))
    for i in tqdm.trange(1, ntraces):
        trace = traces[i].astype("double")
        corr = scipy.signal.correlate(reference, trace)

        max_idxs = corr.argsort()[-10:][::-1]
        max_corr = 0
        max_idx = -1
        for idx in max_idxs:
            lag = lags[idx]
            if lag == 0:
                s_ref = reference
                s_trs = trace
            elif lag < 0:
                s_ref = reference[:lag]
                s_trs = trace[-lag:]
            else:
                s_ref = reference[lag:]
                s_trs = trace[:-lag]
            pcorr = numpy.corrcoef(s_ref, s_trs)[1,0]
            if pcorr > max_corr:
                max_corr = pcorr
                max_idx = idx

        lag = lags[max_idx]
        alignments[i] = lag

    alignments -= numpy.min(alignments)
    max_start = max(alignments)
    min_end = 700

    traces = numpy.zeros((ntraces, 800), 'double')
    for i in range(ntraces):
        traces[i, alignments[i]:alignments[i]+700] = db["traces"][i]

    return traces[:, max_start:min_end]


# P - (n,m) array of n predictions for each of the m candidates
# O - (n,t) array of n traces with t samples each
# returns an (m,t) correlation matrix of m traces t samples each
# TODO : is opt_einsum faster?
class Correlator:
    def __init__(self, P):
        self.P = P - numpy.mean(P, axis=0)
        self.tmp1 = numpy.einsum("nm,nm->m", self.P, self.P, optimize='optimal')
        self.P = self.P.transpose()

    def corr_submatrix(self, O):
        O -= numpy.mean(O, axis=0)
        numerator = self.P @ O
        tmp2 = numpy.einsum("nt,nt->t", O, O, optimize='optimal')
        denominator = numpy.sqrt(numpy.outer(self.tmp1, tmp2))
        return numerator / denominator


def attack_firstorder(db, traces, ntraces):
    logger.info("making first order model")
    model = numpy.repeat(numpy.arange(256, dtype='uint8'), ntraces).reshape((256, ntraces))
    for i in tqdm.trange(ntraces):
        (pt, ct, key, mask, desync) = db["metadata"][i]
        pt_v = pt[2]
        mask_v = mask[15]
        model[:, i] = AES_SBOX[model[:, i] ^ pt_v] ^ mask_v
    model = numpy_popcount.popcount(model).astype("double").transpose()

    # boring, slow, uses too much RAM
    # input_matrix = numpy.vstack((model, traces.transpose()))
    # coefs = numpy.corrcoef(input_matrix)[0:256, 256:]

    # uwu
    correlator = Correlator(model)
    coefs = correlator.corr_submatrix(traces)

    # plot absmax
    coefs = numpy.abs(coefs)
    max_by_key = numpy.max(coefs, axis=1)
    plt.plot(max_by_key)
    plt.show()


def attack_firstorder_fft(db, traces, ntraces):
    # lol
    ntraces = 2000
    db = h5py.File("./ASCAD_databases/ASCAD_desync100.h5", "r")["Attack_traces"]
    traces = db["traces"][0:ntraces, :].astype("double")

    logger.info("making first order model")
    model = numpy.repeat(numpy.arange(256, dtype='uint8'), ntraces).reshape((256, ntraces))
    for i in tqdm.trange(ntraces):
        (pt, ct, key, mask, desync) = db["metadata"][i]
        pt_v = pt[2]
        mask_v = mask[15]
        model[:, i] = AES_SBOX[model[:, i] ^ pt_v] ^ mask_v
    model = numpy_popcount.popcount(model).astype("double").transpose()

    logger.info("making fft traces")
    fft_traces = numpy.zeros(traces.shape, traces.dtype)
    for i in range(traces.shape[0]):
        fft_traces[i, :] = numpy.abs(scipy.fft.fft(traces[i, :]))

    traces = fft_traces[:, 1:350]
    avg = numpy.mean(traces, axis=0)
    peaks = avg.argsort()[-50:][::-1]
    traces = numpy.hstack([traces[:, i:i+1] for i in peaks])

    correlator = Correlator(model)
    coefs = correlator.corr_submatrix(traces)

    # plot absmax
    coefs = numpy.abs(coefs)
    max_by_key = numpy.max(coefs, axis=1)
    plt.plot(max_by_key)
    plt.show()


# def attack_secondorder_fft(db, traces, ntraces):
#     # needs work
#     # lol
#     ntraces = 10000
#     db = h5py.File("./ASCAD_databases/ASCAD.h5", "r")["Attack_traces"]
#     traces = db["traces"][0:ntraces, :].astype("double")
#
#     logger.info("making second order model")
#     model = numpy.repeat(numpy.arange(256, dtype='uint8'), ntraces).reshape((256, ntraces))
#     for i in tqdm.trange(ntraces):
#         (pt, ct, key, mask, desync) = db["metadata"][i]
#         pt_v = pt[2]
#         mask_v = mask[15]
#         model[:, i] = AES_SBOX[model[:, i] ^ pt_v] ^ mask_v
#     model = numpy_popcount.popcount(model).astype("double").transpose()
#
#     logger.info("making fft traces")
#     fft_traces = numpy.zeros((traces.shape[0], traces.shape[1]), traces.dtype)
#     for i in range(traces.shape[0]):
#         fft_i = numpy.abs(scipy.fft.fft(traces[i, :]))
#         conv_i = numpy.abs(numpy.conj(fft_i) * fft_i)
#         fft_traces[i, :] = conv_i
#
#     traces = fft_traces
#     # traces = fft_traces[:, 1:350]
#     # avg = numpy.mean(traces, axis=0)
#     # peaks = avg.argsort()[-50:][::-1]
#     # traces = numpy.hstack([traces[:, i:i+1] for i in peaks])
#
#     correlator = Correlator(model)
#     coefs = correlator.corr_submatrix(traces)
#
#     # plot absmax
#     coefs = numpy.abs(coefs)
#     max_by_key = numpy.max(coefs, axis=1)
#     plt.plot(max_by_key)
#     plt.show()


def attack_secondorder(db, traces, ntraces):
    logger.info("making second order model")
    model = numpy.repeat(numpy.arange(256, dtype='uint8'), ntraces).reshape((256, ntraces))
    for i in tqdm.trange(ntraces):
        (pt, ct, key, mask, desync) = db["metadata"][i]
        pt_v = pt[2]
        model[:, i] = AES_SBOX[model[:, i] ^ pt_v]
    model = numpy_popcount.popcount(model).astype("double").transpose()

    logger.info("making second order combinations")
    nticks = traces.shape[1]
    traces_x = numpy.zeros((ntraces, nticks*(nticks - 1)//2), 'double')
    tx_i = 0
    for i in tqdm.trange(nticks):
        for j in range(i+1, nticks):
            traces_x[:, tx_i] = numpy.abs(traces[:, i] - traces[:, j])
            tx_i += 1

    logger.info("doing corr")
    correlator = Correlator(model)
    ncolumns = 4
    step = traces_x.shape[1] // ncolumns
    all_coefs = []
    for i in tqdm.trange(ncolumns):
        start = i * step
        end = (i + 1) * step
        coefs = correlator.corr_submatrix(traces_x[:, start:end])

        all_coefs.append(coefs)

    all_coefs = numpy.hstack(all_coefs)

    all_coefs = numpy.abs(all_coefs)
    max_by_key = numpy.max(all_coefs, axis=1)
    plt.plot(max_by_key)
    plt.show()

def main(db):
    ntraces = 2000

    # traces = None
    # attack_firstorder_fft(db, traces, ntraces)
    # return

    traces = align(db, ntraces)

    # for i in range(10):
    #     plt.plot(traces[i], color=numpy.hstack((numpy.random.random(3), [0.5])))
    # plt.show()

    attack_firstorder(db, traces, ntraces)
    attack_secondorder(db, traces, ntraces)


if __name__ == "__main__":
    from tqdm.contrib.logging import logging_redirect_tqdm
    logging.basicConfig(level=logging.INFO)
    with logging_redirect_tqdm():
        db = h5py.File("./ASCAD_databases/ASCAD_desync100.h5", "r")
        main(db["Attack_traces"])
